using Microsoft.Win32;
using System;
using System.Globalization;
using System.Threading;
using System.Collections;
using System.Collections.Generic;
using System.Timers;
using System.Diagnostics;

// this application utilizes the Console features of
// intercepting keys, getting a key as it's pressed,
// coloring, and moving buffer regions to make a space
// invaders application. It also takes advantage of the
// feature of generics for any collections it uses
namespace Microsoft.Samples.SpaceInvaders
{
	// an enum for the direction and position in which
	// the players ship is moving
	enum ShipPosition
	{
		Initial = 0,
		MoveLeft = 1,
		MoveRight = 2
	}

	// this class represents a logical 'invader',
	// which is basically comprised of its position
	[CLSCompliant(false)]
	public class Invader
	{
		ItemPosition _position;
		int _id;

		public Invader (ItemPosition ip, int id)
		{
			_position = ip;
			_id = id;
		}

		public ItemPosition Position
		{
			get
			{
				return _position;
			}
			set
			{
				_position = value;
			}
		}

		public int ID
		{
			get
			{
				return _id;
			}
			set
			{
				_id = value;
			}
		}

		public bool CanFire (List<Invader> invaders)
		{
			if (_id > 32)
				return true;

			foreach (Invader i in invaders)
			{
				if (i.ID > _id && ((i.ID - _id) % 8 == 0))
					return false;
			}

			return true;
		}
	}

	// this class represents the player's shields
	// each shield is a single character, and every
	// single shield is tracked independently
	public class Shield
	{
		ItemPosition _position;
		bool _shieldFull;


		public Shield (ItemPosition ip)
		{
			_position = ip;
			_shieldFull = true;
		}

		public ItemPosition Position
		{
			get
			{
				return _position;
			}
			set
			{
				_position = value;
			}
		}

		public bool ShieldFull
		{
			get
			{
				return _shieldFull;
			}
			set
			{
				_shieldFull = value;
			}
		}
	}

	// this class represents the position of an entity,
	// and can include a the position of a shield, a bullet,
	// or an invader
	public class ItemPosition
	{
		private int _left;
		private int _top;

		public ItemPosition (int left, int top)
		{
			_left = left;
			_top = top;
		}

		public int Left
		{
			get
			{
				return _left;
			}
			set
			{
				_left = value;
			}
		}

		public int Top
		{
			get
			{
				return _top;
			}
			set
			{
				_top = value;
			}
		}
	}

	// this class represents the large characters printed
	// on the start screen, and when a game is over
	public class Letter
	{
		ConsoleColor _color;
		string[] _values;
		string _id;

		public Letter(string id, ConsoleColor color, string[] values )
		{
			_color = color;
			_values = values;
			_id = id;
		}

		public string[] GetValues()
		{
			return _values;
		}

		public string ID
		{
			get { return _id; }
		}

		public ConsoleColor Color
		{
			get
			{
				return _color;
			}
		}
	}

	[CLSCompliant(false)]
	public class Invaders
	{
		// collections
		SortedDictionary<int,string[]> highScores;
		List<ItemPosition> liShipBullets;
		List<ItemPosition> liInvaderBullets;
		List<Invader> liInvaders;
		List<Shield> liShieldPositions;
		string[] scoreNames;
		ConsoleColor[] invaderRowColors;
		Dictionary<string, Letter> letters;

		// contstants
		private const char halfShield = (char)9617;
		private const char fullShield = (char)9608;
		private const int initCharVal = 65;
		const int initInvaderTime = 850;

		// events
		System.Timers.Timer shipBulletTimer = new System.Timers.Timer (70);
		System.Timers.Timer HighScoreTimer = new System.Timers.Timer(14000);
		System.Timers.Timer shipDestroyedTimer = new System.Timers.Timer (250);
		System.Timers.Timer explosionTimer = new System.Timers.Timer (35);
		System.Timers.Timer invadersTimer = new System.Timers.Timer ();
		System.Timers.Timer invadersBulletTimer = new System.Timers.Timer (90);

		// variables
		int score = 0;
		int maxInvaderLeft;
		int minInvaderLeft;
		bool invadersMoveRight;
		int iterationCount = 0;
		ItemPosition currentShipPosition;
		ItemPosition invaderExplosion;
		int invaderExplosionState = 0;
		int shipCount = 3;
		bool highScoreDisplayed = false;
		bool shipBeingDestroyed = false;
		bool gettingName = false;
		bool gameOver = true;
		bool gameOverDisplayed = true;
		bool invadersReset = false;
		int shipExplosionState = 0;

		/// </summary>
		[STAThread]
		static void Main (string[] args)
		{
			Invaders id = new Invaders ();
			id.highScores = id.GetScores();

			id.Initialize ();
			id.StartScreen();

			// this loop captures the keys pressed when
			// the game is waiting to be played
			// it only supports viewing the main and highscore
			// screens, or exitting the application
			while (true)
			{
				ConsoleKeyInfo cki = Console.ReadKey (true);

				switch (cki.Key)
				{
					case (ConsoleKey.Escape):
						if (id.highScoreDisplayed)
						{
							id.StartScreen();
							break;
						}
						else
						{
							// exit game
							Console.Clear();
							Console.ForegroundColor = ConsoleColor.Gray;
							Console.CursorVisible = true;
							// write out the high scores
							id.SetScores();
							return;
						}

					case (ConsoleKey.S):
						// start game
						id.gameOver = false;
						id.gameOverDisplayed = false;
						id.StartGame ();
						id.Play ();
						while (id.gettingName) { }
						id.StartScreen();
						break;

					case (ConsoleKey.H) :
						// Display the high score table
						id.ShowHighScores();
						break;

					case (ConsoleKey.M) :
						// Display the high score table
						id.StartScreen();
						break;
				}
			}
		}

		void ShowHighScores()
		{
			// get the high scores first: make sure we can retrieve them ...
			HighScoreTimer.Enabled = false;
			Console.Clear();
			lock (this)
			{
				Console.SetCursorPosition(15, 4);
				WriteEntry("H  ", ConsoleColor.DarkYellow);
				WriteEntry("I  ", ConsoleColor.Magenta);
				WriteEntry("G  ", ConsoleColor.DarkGreen);
				WriteEntry("H              ", ConsoleColor.Cyan);
				WriteEntry("S  ", ConsoleColor.Blue);
				WriteEntry("C  ", ConsoleColor.DarkCyan);
				WriteEntry("O  ", ConsoleColor.Green);
				WriteEntry("R  ", ConsoleColor.DarkMagenta);
				WriteEntry("E  ", ConsoleColor.Red);
				WriteEntry("S  ", ConsoleColor.Cyan);
				Console.SetCursorPosition(12, 7);
				for (int i = 0; i < 10; i++)
				{
					Console.SetCursorPosition(12, 7 + (i * 2));
					WriteEntry(String.Format("{0,8}", scoreNames[i]) + "................................" + String.Format("{0}  {1:00000}", GetVal(i + 1, 0), Convert.ToInt32(GetVal(i + 1, 1), null)), invaderRowColors[i]);
				}
			}
			highScoreDisplayed = true;
			HighScoreTimer.Enabled = true;
		}

		string GetVal(int element, int index)
		{
			return this.highScores[element][index];
		}

		// retrieves the scores from the registry,
		// supplying default scores if the game hasn't
		// been played yet. I ignore any errors, and simply
		// return the default set of scores if something
		// goes wrong reading the Registry
		SortedDictionary<int,string[]> GetScores()
		{
			SortedDictionary<int, string[]> scores = new SortedDictionary<int, string[]>();
			RegistryKey rk = null;

			try
			{
				rk = Registry.LocalMachine.CreateSubKey(@"Software\Demos\ConsoleInvaders");

				for (int i = 0; i < 10; i++)
				{
					string s = ((char)(initCharVal + i)).ToString(CultureInfo.InvariantCulture);
					scores.Add(i + 1, (string[])rk.GetValue("Score" + (i + 1).ToString(CultureInfo.InvariantCulture), new string[] { s + s + s, (5000 - (i * 500)).ToString(CultureInfo.InvariantCulture) }, false));
				}
			}
			catch (Exception)
			{
				for (int i = 0; i < 10; i++)
				{
					string s = ((char)(initCharVal + i)).ToString(CultureInfo.InvariantCulture);
					scores[i + 1] = new string[] { s + s + s, (5000 - (i * 500)).ToString(CultureInfo.InvariantCulture) };
				}
			}
			finally
			{
				rk.Close();
			}

			return scores;
		}

		// writes the scores out to the registry
		// I ignore any errors that occur because
		// this functionality is not critical to the game
		// all this means is that if it DOES fail,
		// then your scores won't be saved
		void SetScores()
		{
			RegistryKey rk = null;
			try
			{
				rk = Registry.LocalMachine.OpenSubKey(@"Software\Demos\ConsoleInvaders", true);

				for (int i = 1; i < 11; i++)
				{
					rk.SetValue("Score" + i.ToString(CultureInfo.InvariantCulture), this.highScores[i]);
				}
			}
			catch (Exception)
			{
				// can't tell user it failed: don't worry about it ...
			}
			finally
			{
				rk.Close();
			}
		}

		// figures out if the score for the current player
		// is good enough to make the high score table
		// if it is, then the score is written to
		// the high score table, having been entered
		// by the user
		void WriteScore(int scoreToWrite)
		{
			lock (this)
			{
				int valToWriteOver = 0;

				foreach (KeyValuePair<int, string[]> kvp in this.highScores)
				{
					if (scoreToWrite > Convert.ToInt32(kvp.Value[1], null))
					{
						valToWriteOver = kvp.Key;
						break;
					}
				}

				if (valToWriteOver > 0)
				{
					Console.Clear();
					lock (this)
					{
						Console.SetCursorPosition(15, 6);
						Console.ForegroundColor = ConsoleColor.Cyan;
						Console.Write("You made the high score table!");
						Console.SetCursorPosition(15, Console.CursorTop + 2);
						Console.Write("Please enter your initals ==> AAA");
						Console.SetCursorPosition(Console.CursorLeft - 3, Console.CursorTop);
					}
					string initials = CaptureInitials();


					for (int i = 10; i > valToWriteOver; i--)
					{
						this.highScores[i] = this.highScores[i - 1];
					}

					this.highScores[valToWriteOver] = new String[] { initials, scoreToWrite.ToString(CultureInfo.InvariantCulture) };
				}
			}
			
		}

		// captures the initials of the user for the
		// high score table
		string CaptureInitials()
		{
			Char curA = (char)initCharVal;
			Char curB = (char)initCharVal;
			Char curC = (char)initCharVal;
			int position = 0;
			gettingName = true;

			while (true)
			{
				Console.CursorSize = 100;
				Console.CursorVisible = true;
				ConsoleKeyInfo cki = Console.ReadKey(true);

				switch (cki.Key) {
					case ConsoleKey.DownArrow:
						if (position == 0)
						{
							SetValUp(ref curA);
						}
						else if (position == 1)
						{
							SetValUp(ref curB);
						}
						else
						{
							SetValUp(ref curC);
						}
						Console.SetCursorPosition(Console.CursorLeft - 1, Console.CursorTop);
						break;
					case ConsoleKey.UpArrow:
						if (position == 0)
						{
							SetValDown(ref curA);
						}
						else if (position == 1)
						{
							SetValDown(ref curB);
						}
						else
						{
							SetValDown(ref curC);
						}
						Console.SetCursorPosition(Console.CursorLeft - 1, Console.CursorTop);
						break;

					case ConsoleKey.LeftArrow:
						if (position > 0)
						{
							position = position - 1;
							Console.SetCursorPosition(Console.CursorLeft - 1, Console.CursorTop);
						}
						else
						{
							position = position + 2;
							Console.SetCursorPosition(Console.CursorLeft + 2, Console.CursorTop);
						}
						break;

					case ConsoleKey.RightArrow:
						if (position < 2)
						{
							position = position + 1;
							Console.SetCursorPosition(Console.CursorLeft + 1, Console.CursorTop);
						}
						else
						{
							position = position - 2;
							Console.SetCursorPosition(Console.CursorLeft - 2, Console.CursorTop);
						}
						break;
					case ConsoleKey.SpaceBar:
						if (position == 1)
						{
							Console.Write(" ");
							position++;
						}
						break;
					case ConsoleKey.BackSpace :
						if (position > 0)
						{
							Console.SetCursorPosition(Console.CursorLeft - 1, Console.CursorTop);
							position--;
						}
						break;
					case ConsoleKey.Enter:
						if (position == 2)
						{
							Console.CursorVisible = false;
							return curA.ToString(CultureInfo.InvariantCulture) + curB.ToString(CultureInfo.InvariantCulture) + curC.ToString(CultureInfo.InvariantCulture);
						}
						else
						{
							Console.SetCursorPosition(Console.CursorLeft + 1, Console.CursorTop);
							position++;
						}
						break;
				}

				if (cki.Key >= ConsoleKey.A && cki.Key <= ConsoleKey.Z)
				{
					Console.Write(cki.KeyChar);

					if (position == 2)
					{
						Console.SetCursorPosition(Console.CursorLeft - 1, Console.CursorTop);
						curC = cki.KeyChar;
					}
					else
					{
						if (position == 0)
							curA = cki.KeyChar;
						else
							curB = cki.KeyChar;
						position++;
					}
				}
			}
		}

		// determines if the user has entered an entry
		// which means that the initial they are 
		// entering should wrap back to the original
		void SetValUp(ref char curVal)
		{
			if (curVal < (char)(65 + 25))
			{
				curVal++;
			}
			else
			{
				curVal = (char)65;
			}

			Console.Write(curVal);
		}

		// determines if the user has entered an entry
		// which means that the initial they are 
		// entering should wrap back to the original
		void SetValDown(ref char curVal)
		{
			if (Convert.ToInt32(curVal, null) > 65)
			{
				curVal--;
			}
			else
			{
				curVal = (char)(65 + 25);
			}

			Console.Write(curVal);
		}

		// The main code for the play part of the program
		// This routine captures key presses from the user
		// in order to move the ship. Only certain keys
		// are supported, many are ignored.
		// As this routine is actioning, the timers for
		// moving the invaders, generating bullets, 
		// and determining collisions occur, but
		// not as a direct result of this code
		void Play ()
		{
			DrawShip (ShipPosition.Initial);
			while (true)
			{
				// the true parameter indicates
				// that the key press should not be sent
				// to the main window of the console,
				// but intercepted, and just sent here
				ConsoleKeyInfo cki = Console.ReadKey (true);

				if (gameOver || gettingName)
					return;

				switch (cki.Key)
				{
					case (ConsoleKey.LeftArrow) :
						// move the ship left ...
						if (!shipBeingDestroyed)
						{
							lock (this)
							{
								DrawShip(ShipPosition.MoveLeft);
							}
						}

						break;

					case (ConsoleKey.RightArrow) :
						// move the ship right ...
						if (!shipBeingDestroyed)
						{
							lock (this)
							{
								DrawShip(ShipPosition.MoveRight);
							}
						}

						break;

					case (ConsoleKey.SpaceBar) :
						// bullet was fired!
						if (!shipBeingDestroyed)
						{
							lock (this)
							{
								ShootBullet();
							}
						}
						break;
					case (ConsoleKey.Escape) :
						Console.Clear();
						DisableTimers(false);
						ClearLists();
						return;

					case (ConsoleKey.P):
						// loop here until p pressed again ...
						DisableTimers(false);
						while (true)
						{
							ConsoleKeyInfo ckNext = Console.ReadKey(true);
							if (ckNext.Key == ConsoleKey.P)
							{
								break;
							}
						}
						DisableTimers(true);			
						break;
				}
			}
		}

		// resets the lists
		void ClearLists ()
		{
			liShipBullets = new List<ItemPosition>();
			liShieldPositions = new List<Shield>();
			liInvaders = new List<Invader>();
			liInvaderBullets = new List<ItemPosition>();
		}

		// a standard routine for chaging the color, 
		// and writing something to the Console
		void WriteEntry (string entry, ConsoleColor foreColor)
		{
			Console.ForegroundColor = foreColor;
			Console.Write (entry);
		}

		// the routine to run the very first time
		// the application starts. The positioning,
		// and sizing code is simply to ensure the window
		// size is viable for the game
		void Initialize ()
		{
			Console.BackgroundColor = ConsoleColor.Black;
			Console.SetWindowPosition(0,0);
			if (Console.WindowHeight > 72 || Console.WindowWidth > 36)
                Console.SetWindowSize (10, 10);
			Console.SetBufferSize(73, 37);
			Console.SetWindowSize (Console.BufferWidth - 1, Console.BufferHeight - 1);
			Console.CursorVisible = false;
			Console.Title = "Managed Space Invaders";
			HighScoreTimer.Elapsed += new ElapsedEventHandler(OnHighScoreTimedEvent);
			shipBulletTimer.Elapsed += new ElapsedEventHandler(OnBulletTimedEvent);
			invadersTimer.Elapsed += new ElapsedEventHandler(OnInvadersTimedEvent);
			invadersBulletTimer.Elapsed += new ElapsedEventHandler(OnInvadersBulletTimedEvent);
			explosionTimer.Elapsed += new ElapsedEventHandler(OnInvadersExplosion);
			shipDestroyedTimer.Elapsed += new ElapsedEventHandler(OnShipDestroyed);

			scoreNames = new string[10];
			scoreNames[0] = "First";
			scoreNames[1] = "Second";
			scoreNames[2] = "Third";
			scoreNames[3] = "Fourth";
			scoreNames[4] = "Fifth";
			scoreNames[5] = "Sixth";
			scoreNames[6] = "Seventh";
			scoreNames[7] = "Eighth";
			scoreNames[8] = "Ninth";
			scoreNames[9] = "Tenth";

			invaderRowColors = new ConsoleColor[12];
			invaderRowColors[0] = ConsoleColor.Red;
			invaderRowColors[1] = ConsoleColor.White;
			invaderRowColors[2] = ConsoleColor.Gray;
			invaderRowColors[3] = ConsoleColor.Yellow;
			invaderRowColors[4] = ConsoleColor.Green;
			invaderRowColors[5] = ConsoleColor.Cyan;
			invaderRowColors[6] = ConsoleColor.DarkYellow;
			invaderRowColors[7] = ConsoleColor.DarkGreen;
			invaderRowColors[8] = ConsoleColor.Blue;
			invaderRowColors[9] = ConsoleColor.Magenta;
			invaderRowColors[10] = ConsoleColor.Red;
			invaderRowColors[11] = ConsoleColor.DarkRed;

			MakeLetters();
		}

		// makes the large letters displayed on the start
		// screen, and on the game over screen
		void MakeLetters()
		{
			letters = new Dictionary<string,Letter>();
			letters.Add("S", new Letter("S", ConsoleColor.White, new string[] { "SSSSS", "S   S", "S", "SSSSS", "    S", "S   S", "SSSSS" }));
			letters.Add("P", new Letter("P", ConsoleColor.Green, new string[] { "PPPPP", "P   P", "P   P", "PPPPP", "P", "P", "P" }));
			letters.Add("A", new Letter("A", ConsoleColor.Red, new string[] { "  A", " AAA", " A A", "AA AA", "AAAAA", "A   A", "A   A" }));
			letters.Add("C", new Letter("C", ConsoleColor.Gray, new string[] { "CCCCC", "C   C", "C", "C", "C", "C   C", "CCCCC" }));
			letters.Add("E", new Letter("E", ConsoleColor.Blue, new string[] { "EEEEE", "E", "E", "EEE", "E", "E", "EEEEE" }));
			letters.Add("I", new Letter("I", ConsoleColor.Blue, new string[] { "IIIII", "  I", "  I", "  I", "  I", "  I", "IIIII" }));
			letters.Add("N", new Letter("N", ConsoleColor.DarkYellow, new string[] { "N   N", "NN  N", "NN  N", "N N N", "N  NN", "N  NN", "N   N" }));
			letters.Add("V", new Letter("V", ConsoleColor.Yellow, new string[] { "V   V", "V   V", "V   V", "V   V", "V   V", " V V", "  V" }));
			letters.Add("D", new Letter("D", ConsoleColor.Magenta, new string[] { "DDD", "D  D", "D   D", "D   D", "D   D", "D  D", "DDD" }));
			letters.Add("R", new Letter("R", ConsoleColor.Cyan, new string[] { "RRRR", "R  R", "R   R", "RRRRR", "R   R", "R   R", "R   R" }));
			letters.Add("O", new Letter("O", ConsoleColor.Gray, new string[] { "OOOOO", "O   O", "O   O", "O   O", "O   O", "O   O", "OOOOO" }));
			letters.Add("G", new Letter("G", ConsoleColor.Blue, new string[] { "GGGGG", "G   G", "G", "G", "G  GG", "G   G", "GGGGG" }));
			letters.Add("M", new Letter("M", ConsoleColor.White, new string[] { "M   M", "MM MM", "M M M", "M M M", "M   M", "M   M", "M   M" }));
		}

		// resets the game, which occurs when the user
		// destroys all the current invaders
		void Reset()
		{
			DisableTimers(false);
			ResetTimers();
		}

		// this routine is used to draw all the letters
		// provided by the user, such as 'space invaders'
		// or, 'game over'. If you want to write multiple
		// rows of letters, call this routine multiple times,
		// since it can only write letters on a single row
		void DrawLetters(int left, int top, int leftAddition, int sleep, params string[] rows)
		{
			int count = 0;
			int leftPos = left;

			foreach (string row in rows)
			{
				lock (this)
				{
					foreach (string s in letters[row].GetValues())
					{
						Console.SetCursorPosition(leftPos, top + count);
						WriteEntry(s, letters[row].Color);
						count++;
					}
				}
				if (sleep > 0)
					Thread.Sleep(sleep);

				leftPos += leftAddition;
				count = 0;
			}
		}

		// draws the start screen for the application
		void StartScreen ()
		{
			Console.Clear();
			// space
			DrawLetters(6, 2, 12, 0, new string[] { "S", "P", "A", "C", "E" });

			// invaders
			DrawLetters(6, 10, 7, 0, new string[] { "I", "N", "V", "A", "D", "E", "R", "S" });

			Console.SetCursorPosition (4, 20);
			WriteEntry ("Esc    End Game", ConsoleColor.Cyan);
			Console.SetCursorPosition (4, 22);
			WriteEntry ("S      Start New Game", ConsoleColor.White);
			Console.SetCursorPosition (4, 24);
			WriteEntry ("P      Pause (In Game)", ConsoleColor.Green);
			Console.SetCursorPosition (4, 26);
			WriteEntry ("H      High Score Table", ConsoleColor.Red);

			Console.SetCursorPosition (4, 20);

			// this loop simply prints out the colors of
			// each invader, and their value, so the user
			// knows what the invaders are worth.
			// Basic concept for score is, the lower
			// the invader, the more danger to the user,
			// therefore the higher the points
			for (int i = 1; i < 12; i++)
			{
				ConsoleColor color = invaderRowColors[i];
				DrawInvader (46, i + 19, color);
				Console.SetCursorPosition (51, i + 19);
				WriteEntry (" = " + String.Format ("{0:00}", (i * 5)) + " points", color);
			}
			highScoreDisplayed = false;
		}

		// initialization that occurs for the game,
		// each time a new game is started
		void StartGame ()
		{
			Console.Clear ();
			iterationCount = 0;
			ClearLists();
			Reset();

			// other variables to initialize ...
			highScoreDisplayed = false;
			shipBeingDestroyed = false;
			invadersMoveRight = true;
			shipCount = 3;
			iterationCount = 0;

			// Score
			ScoreUpdate ();

			// Ships
			DrawReserveShips ();
			InitializeInvaders (iterationCount);
			iterationCount = iterationCount + 1;
			invadersBulletTimer.Enabled = true;

			// shields
			DrawShields ();
		}

		// these are the extra lives for the user,
		// which are drawn in the top right of the screen
		void DrawReserveShips ()
		{
			Console.ForegroundColor = ConsoleColor.Cyan;
			for (int i = 2; i > 0; i--)
				Console.MoveBufferArea(0, 0, 3, 1, Console.WindowWidth - 3 - (3 * i), 0);

			for (int i = shipCount - 1; i > 0; i--)
			{
				lock (this)
				{
					Console.SetCursorPosition(Console.WindowWidth - 3 - (3 * i), 0);
					Console.Write("{0}{1}{2}", (char)9568, (char)9575, (char)9571);
				}
			}
		}

		// simply draws all of the shields on the screen,
		// and populates the list keeping track of the
		// shield state. Right now, shields have two states,
		// full, or half destroyed. You could remove
		// the half destroyed concept if you wanted
		void DrawShields ()
		{
			liShieldPositions = new List<Shield>();
			Console.ForegroundColor = ConsoleColor.Gray;

			for (int i = 0; i < 5; i++)
			{
				Console.SetCursorPosition(i * 14 + 5, Console.WindowHeight - 8);
				Console.Write("{0}{0}{0}{0}{0}{0}", fullShield);
				Console.SetCursorPosition(i * 14 + 5, Console.WindowHeight - 7);
				Console.Write("{0}{0}  {0}{0}", fullShield);
				Console.SetCursorPosition(i * 14 + 5, Console.WindowHeight - 6);
				Console.Write("{0}{0}  {0}{0}", fullShield);

				for (int j = 5; j <= 10; j++)
				{
					liShieldPositions.Add(new Shield(new ItemPosition(i * 14 + j, Console.WindowHeight - 8)));
				}

				for (int j = 0; j <= 1; j++)
				{
					liShieldPositions.Add(new Shield(new ItemPosition(i * 14 + 5, Console.WindowHeight - (7 - j))));
					liShieldPositions.Add(new Shield(new ItemPosition(i * 14 + 6, Console.WindowHeight - (7 - j))));
					liShieldPositions.Add(new Shield(new ItemPosition(i * 14 + 9, Console.WindowHeight - (7 - j))));
					liShieldPositions.Add(new Shield(new ItemPosition(i * 14 + 10, Console.WindowHeight - (7 - j))));
				}
			}
		}

		// resolves any collisions with the shields
		// from bullets, either from the player, or from
		// the invaders. Note that when the invaders
		// drop below the line of the shields, the shields
		// are simply destroyed outright, there is no
		// collsion
		void ResolveShieldCollisions ()
		{
			foreach (Shield shield in liShieldPositions)
			{
				bool destroyed = false;

				// resolve invader bullet collisions
				foreach (ItemPosition ip in liInvaderBullets)
				{
					if (ip.Left == shield.Position.Left && ip.Top == shield.Position.Top)
					{
						if (shield.ShieldFull)
						{
							lock (this)
							{
								Console.SetCursorPosition(shield.Position.Left, shield.Position.Top);
								Console.ForegroundColor = ConsoleColor.Gray;
								Console.Write(halfShield);
								shield.ShieldFull = false;
							}
						}
						else
						{
							Console.MoveBufferArea (0, 0, 1, 1, ip.Left, ip.Top);
							liShieldPositions.Remove (shield);
						}
						liInvaderBullets.Remove (ip);
						break;
					}
				}

				if (!destroyed)
				{
					// resolve player bullet collisions
					foreach (ItemPosition ip in liShipBullets)
					{
						if (ip.Left == shield.Position.Left && ip.Top == shield.Position.Top)
						{
							if (shield.ShieldFull)
							{
								lock (this)
								{
									Console.SetCursorPosition(shield.Position.Left, shield.Position.Top);
									Console.ForegroundColor = ConsoleColor.Gray;
									Console.Write(halfShield);
									shield.ShieldFull = false;
								}
							}
							else
							{
								Console.MoveBufferArea (0, 0, 1, 1, ip.Left, ip.Top);
								liShieldPositions.Remove (shield);
							}
							liShipBullets.Remove (ip);
							break;
						}
					}
				}
			}
		}

		// draws the invaders for the first time
		// the top row increases with each iteration
		// of the game, making the invaders start on a 
		// lower, and lower level. There currently is no
		// minimum row (an imporovement yet to make)
		void InitializeInvaders (int topRow)
		{
			invadersTimer.Enabled = false;
			invadersTimer.Interval = initInvaderTime;
			liInvaders = new List<Invader>();
			invadersMoveRight = false;

			int initCursorLeft = Console.CursorLeft;
			int initCursorTop = Console.CursorTop;
			int invaderCount = 0;

			Console.BackgroundColor = ConsoleColor.Black;
			for (int j = topRow; j < 5 + topRow; j++)
			{
				ConsoleColor foreColor = invaderRowColors[j + 1];

				for (int i = 0; i < 8; i++)
				{
					int k = i * 7 + 5;
					int l = (j * 3 + 2) + 3;

					Console.SetCursorPosition (k, l);
					invaderCount++;
					liInvaders.Add (new Invader (new ItemPosition (k, l), invaderCount));

					Console.ForegroundColor = foreColor;
					Console.Write ("{0}{1}{2}", (char)9561, (char)9573, (char)9564);
				}
			}

			Console.BackgroundColor = ConsoleColor.Black;
			Console.SetCursorPosition (initCursorLeft, initCursorTop);
			invadersTimer.Enabled = true;
		}

		// simply shifts an invader to its new position
		// based on the timer having moved the invader
		void DrawInvader (int left, int top, ConsoleColor foreColor)
		{
			Console.SetCursorPosition (left, top);
			Console.ForegroundColor = foreColor;
			Console.Write ("{0}{1}{2}", (char)9561, (char)9573, (char)9564);
		}

		// this routine randomly generates invader bullets
		// You will notice that over time, the likelihood
		// of a bullet being generated increases, meaning
		// that as more levels are reached, the invaders
		// fire more often, making the game harder
		void GenerateInvaderBullet ()
		{
			Random r = new Random (DateTime.Now.Millisecond);
			int next = r.Next (1000);

			if (next > (1000 - 40 * iterationCount))
			{
				List<Invader> InvadersCanFire = new List<Invader>();

				// identify a random invader, and have it shoot
				foreach (Invader ipInvader in liInvaders)
				{
					if (ipInvader.CanFire (liInvaders))
						InvadersCanFire.Add (ipInvader);
				}

				int nextInvader = r.Next (InvadersCanFire.Count);

				// for now, just make the last one fire!
				Invader invader = InvadersCanFire[nextInvader];

				liInvaderBullets.Add (new ItemPosition (invader.Position.Left + 1, invader.Position.Top + 1));
				lock (this)
				{
					Console.SetCursorPosition(invader.Position.Left + 1, invader.Position.Top + 1);
					Console.ForegroundColor = ConsoleColor.White;
					Console.Write((char)9553);
				}
			}
		}


		// draw an invader at a specific position
		void DrawInvaderAt (Invader ip, ref int invaderMaxLeft, ref int invaderMinLeft)
		{
			if (!invadersMoveRight)
			{
				Console.MoveBufferArea(0, 0, 3, 1, ip.Position.Left, ip.Position.Top);
				DrawInvader(ip.Position.Left + 1, ip.Position.Top, invaderRowColors[(ip.Position.Top - 2) / 3]);
			}
			else
			{
				Console.MoveBufferArea(0, 0, 3, 1, ip.Position.Left, ip.Position.Top);
				DrawInvader(ip.Position.Left - 1, ip.Position.Top, invaderRowColors[(ip.Position.Top - 2) / 3]);
			}

			if (invaderMaxLeft < ip.Position.Left)
				invaderMaxLeft = ip.Position.Left;

			if (invaderMinLeft > ip.Position.Left)
				invaderMinLeft = ip.Position.Left;
		}

		// when the invaders reach the side of the screen,
		// they come down a row. THis routine figures out
		// if it's time for them to move down
		void MoveInvadersDown ()
		{
			for (int i = liInvaders.Count - 1; i >= 0; i--)
			{
				ItemPosition ip = ((Invader)(liInvaders[i])).Position;

				Console.MoveBufferArea(0, 0, 3, 1, ip.Left, ip.Top);
				ip.Top = ip.Top + 3;
				DrawInvader(ip.Left, ip.Top, invaderRowColors[(ip.Top - 2) / 3]);
			}

			// remove any valid entries in the shields
			int temp = liShieldPositions.Count - 1;

			for (int i = temp; i >= 0; i--)
			{
				if (((Invader)liInvaders[liInvaders.Count - 1]).Position.Top >= ((Shield)liShieldPositions[i]).Position.Top)
				{
					Console.MoveBufferArea(0, 0, 1, 1, ((Shield)liShieldPositions[i]).Position.Left, ((Shield)liShieldPositions[i]).Position.Top);
					liShieldPositions.RemoveAt(i);
				}
			}

			if (((Invader)liInvaders[liInvaders.Count - 1]).Position.Top >= currentShipPosition.Top)
			{
				// game Over!
				Thread.Sleep (2500);
				gameOver = true;
				return;
			}

			invadersMoveRight = !invadersMoveRight;
		}

		void RedrawInvaders ()
		{
			int initCursorLeft = Console.CursorLeft;
			int initCursorTop = Console.CursorTop;
			int tempInvaderMaxLeft = 0;
			int tempInvaderMinLeft = Console.WindowWidth;

			if (!invadersMoveRight)
			{
				if (maxInvaderLeft < Console.WindowWidth - 5)
				{
					foreach (Invader i in liInvaders)
					{
						DrawInvaderAt (i, ref tempInvaderMaxLeft, ref tempInvaderMinLeft);
						i.Position.Left = i.Position.Left + 1;
					}
				}
				else
				{
					MoveInvadersDown ();
				}
			}
			else
			{
				if (minInvaderLeft > 5)
				{
					foreach (Invader ip in liInvaders)
					{
						DrawInvaderAt (ip, ref tempInvaderMaxLeft, ref tempInvaderMinLeft);
						ip.Position.Left = ip.Position.Left - 1;
					}
				}
				else
				{
					MoveInvadersDown ();
				}
			}

			minInvaderLeft = tempInvaderMinLeft;
			maxInvaderLeft = tempInvaderMaxLeft;
		}

		// this routine occurs when the user fires
		void ShootBullet ()
		{
			// logic is peppered which allows multiple bullets
			// to fire. This code here ensures that in this
			// version, only one does, but we could allow
			// more than 1 in the future
			if (liShipBullets.Count == 1)
			{
				return;
			}

			Console.BackgroundColor = ConsoleColor.Black;
			Console.SetCursorPosition (currentShipPosition.Left + 1, currentShipPosition.Top - 1);
			liShipBullets.Add (new ItemPosition (currentShipPosition.Left + 1, currentShipPosition.Top - 1));
			Console.ForegroundColor = ConsoleColor.Yellow;
			Console.Write ("|");
			Console.ForegroundColor = ConsoleColor.White;

			shipBulletTimer.Enabled = true;
		}

		// redraws the users ship
		void DrawShip (ShipPosition sp)
		{
			if (sp == ShipPosition.Initial)
			{
				Console.SetCursorPosition (Console.WindowWidth / 2 - 1, Console.WindowHeight - 2);
				currentShipPosition = new ItemPosition (Console.CursorLeft, Console.CursorTop);

				Console.ForegroundColor = ConsoleColor.Cyan;
				Console.Write ("{0}{1}{2}", (char)9568, (char)9575, (char)9571);
			}
			else if (sp == ShipPosition.MoveLeft && currentShipPosition.Left > 2)
			{
				Console.MoveBufferArea(currentShipPosition.Left, currentShipPosition.Top, 3, 1, currentShipPosition.Left - 1, currentShipPosition.Top);
				currentShipPosition.Left = currentShipPosition.Left - 1;
				ResolveShipDestroyed ();
			}
			else if (sp == ShipPosition.MoveRight && currentShipPosition.Left < Console.WindowWidth - 4)
			{
				Console.MoveBufferArea(currentShipPosition.Left, currentShipPosition.Top, 3, 1, currentShipPosition.Left + 1, currentShipPosition.Top);
				currentShipPosition.Left = currentShipPosition.Left + 1;
				ResolveShipDestroyed ();
			}
		}

		// updates the score after an invader is downed
		void ScoreUpdate ()
		{
			Console.ForegroundColor = ConsoleColor.White;
			Console.SetCursorPosition (10, 0);
			Console.Write (String.Format ("{0:00000}", score));
		}

		// shoots down an invader, and removes
		// it from the invader list. When all invaders
		// are destroyed, the level is complete, and
		// the next level starts after a delay
		bool ShootDownInvader (ItemPosition bullet, Invader invader)
		{
			if (bullet.Top == invader.Position.Top && bullet.Left >= invader.Position.Left && bullet.Left < invader.Position.Left + 3)
			{
				// reset timer!
				if (liInvaders.Count == 1)
				{
					invadersReset = true;
					invadersTimer.Interval = initInvaderTime;
				}

				//destroyInvader!
				liShipBullets.Remove (bullet);
		
				Console.MoveBufferArea(0, 0, 3, 1, invader.Position.Left, invader.Position.Top);

				int row = ((invader.Position.Top - 2) / 3);
				score += (row + 1) * 5;
				ScoreUpdate ();
				lock (this)
				{
					Console.SetCursorPosition(invader.Position.Left, invader.Position.Top);
					Console.ForegroundColor = ConsoleColor.Cyan;
					Console.Write("{0}{0}{0}", (char)9788);
				}
				invaderExplosion = invader.Position;
				explosionTimer.Enabled = true;
				liInvaders.Remove (invader);
				if (liInvaders.Count == 0)
				{
					Thread.Sleep(500);
					Console.MoveBufferArea(0, 0, 3, 1, invaderExplosion.Left, invaderExplosion.Top);
					Thread.Sleep (3000);
					InitializeInvaders (iterationCount);
					DrawShields();
					iterationCount++;
					invadersReset = false;
					return true;
				}

				// this switch simply speeds the invaders up
				// over time, making them approach the user
				// faster as their numbers are reduced
				switch (Convert.ToInt32(invadersTimer.Interval, null))
				{
					case (initInvaderTime) :
						if (liInvaders.Count < 35)
							invadersTimer.Interval = 700;
						break;
					case (700) :
						if (liInvaders.Count < 28)
							invadersTimer.Interval = 550;
						break;
					case (550):
						if (liInvaders.Count < 20)
							invadersTimer.Interval = 350;
						break;
					case (350):
						if (liInvaders.Count < 10)
							invadersTimer.Interval = 175;
						break;
					case (175):
						if (liInvaders.Count < 5)
							invadersTimer.Interval = 90;
						break;
					case (90):
						if (liInvaders.Count < 3)
							invadersTimer.Interval = 50;
						break;
					case (50):
						if (liInvaders.Count < 2)
							invadersTimer.Interval = 35;
						break;
					default:
						if (liInvaders.Count < 1)
							invadersTimer.Interval = initInvaderTime;
						break;
				}

				return true;
			}

			return false;
		}

		// determines if two bullets collide, and if
		// so, destroys them both
		private void ResolveBulletCollisions ()
		{
			foreach (ItemPosition ipsb in liShipBullets)
			{
				foreach (ItemPosition ipib in liInvaderBullets)
				{
					if (ipsb.Left == ipib.Left)
					{
						if (ipib.Top == ipsb.Top)
						{
							liShipBullets.Remove (ipsb);
							liInvaderBullets.Remove (ipib);
							Console.MoveBufferArea (0, 0, 1, 1, ipsb.Left, ipsb.Top);
							break;
						}
					}
				}
			}
		}

		// determines if an invader is destroyed
		private void InvaderDestroyed ()
		{
			bool invaderShot = false;

			foreach (ItemPosition ipBullet in liShipBullets)
			{
				foreach (Invader ipInvader in liInvaders)
				{
					invaderShot = ShootDownInvader (ipBullet, ipInvader);
					if (invaderShot)
						break;
				}

				if (!invaderShot)
				{
					if (ipBullet.Top <= 3)
					{
						Console.MoveBufferArea (0, 0, 1, 1, ipBullet.Left, ipBullet.Top);
						liShipBullets.Remove (ipBullet);
					}
					else
					{
						Console.MoveBufferArea (ipBullet.Left, ipBullet.Top, 1, 1, ipBullet.Left, ipBullet.Top - 1);
						ipBullet.Top = ipBullet.Top - 1;
					}
				}
			}

			if (liShipBullets.Count == 0)
				shipBulletTimer.Enabled = false;
		}

		// determins if the user is destroyed. If so, 
		// and this is the users last life, then the 
		// game is over. The high score prompt may occur
		// at this point
		void ResolveShipDestroyed ()
		{
			lock (this)
			{
				foreach (ItemPosition ip in liInvaderBullets)
				{
					if (ip.Top == currentShipPosition.Top)
					{
						if (ip.Left >= currentShipPosition.Left && ip.Left <= currentShipPosition.Left + 2)
						{
							shipBeingDestroyed = true;

							// explode ship!
							shipCount -= 1;

							// draw over the existing bullets
							foreach (ItemPosition ipi in liInvaderBullets)
							{
								Console.MoveBufferArea(0, 0, 1, 1, ipi.Left, ipi.Top);
							}

							// reset the invader bullets ...
							liInvaderBullets = new List<ItemPosition>();
							invadersBulletTimer.Enabled = false;
							invadersTimer.Enabled = false;
							shipBulletTimer.Enabled = false;
							shipDestroyedTimer.Enabled = true;
						}
					}
				}
			}
		}

		void DisableTimers (bool enable)
		{
			lock (this)
			{
				invadersBulletTimer.Enabled = enable;
				invadersTimer.Enabled = enable;
				shipBulletTimer.Enabled = enable;

				// these timers should never be enabled, except
				// inline as appropriate
				shipDestroyedTimer.Enabled = false;
				explosionTimer.Enabled = false;
			}
		}

		// resets all timers back to their original values
		void ResetTimers()
		{
			invadersBulletTimer.Interval = 90;
			invadersTimer.Interval = 850;
			shipBulletTimer.Interval = 70;
			shipDestroyedTimer.Interval = 250;
			explosionTimer.Interval = 35;
		}

		// displays the game over screen
		// if the user needs to enter in their name for
		// a high score, they are prompoted at this point
		void DisplayGameOver ()
		{
			if (!gameOverDisplayed)
			{
				Console.Clear ();
				gameOver = true;
				gameOverDisplayed = true;
				ScoreUpdate ();
				WriteScore(score);
				score = 0;
				DrawLetters(20, (Console.WindowHeight / 2) - 8, 8, 350, new string[]{ "G", "A", "M", "E" });
				DrawLetters(20, (Console.WindowHeight / 2), 8, 350, new string[] { "O", "V", "E", "R" });
				Thread.Sleep (3000);
				gettingName = false;
			}
		}

		// This event tracks the movement, and collisions
		// of the players bullet
		private void OnBulletTimedEvent (object source, ElapsedEventArgs e)
		{
			lock (this)
			{
				if (liShipBullets.Count == 0)
				{
					shipBulletTimer.Enabled = false;
					return;
				}

				ResolveBulletCollisions ();
				ResolveShieldCollisions ();
				InvaderDestroyed ();
			}
		}

		// this event moves the invaders on the screen
		private void OnInvadersTimedEvent (object source, ElapsedEventArgs e)
		{
			if (!invadersReset)
			{
				lock (this)
				{
					RedrawInvaders ();
					if (gameOver)
					{
						Reset();
						DisplayGameOver ();
						return;
					}

					InvaderDestroyed ();
				}
			}
		}

		// this event occurs when the player's ship
		// is destroyed. Cycling occurs when their ship
		// is destroyed, to give it an explosion feel.
		// The cycling is simply based on this event
		// firing again. The game may be over based on this
		// event
		private void OnShipDestroyed (object source, ElapsedEventArgs e)
		{
			lock (this)
			{
				if (shipExplosionState < 5)
				{
					Console.SetCursorPosition (currentShipPosition.Left, currentShipPosition.Top);
					if (shipExplosionState == 0)
						Console.ForegroundColor = ConsoleColor.DarkYellow;
					else if (shipExplosionState == 1)
						Console.ForegroundColor = ConsoleColor.Yellow;
					else if (shipExplosionState == 2)
						Console.ForegroundColor = ConsoleColor.Green;
					else if (shipExplosionState == 3)
						Console.ForegroundColor = ConsoleColor.Cyan;
					else if (shipExplosionState == 4)
						Console.ForegroundColor = ConsoleColor.Red;

					Console.Write ("{0}{0}{0}", (char)9788);
					shipExplosionState++;
				}
				else
				{
					if (shipCount == 0)
					{
						Thread.Sleep(1500);
						Reset();
						DisplayGameOver ();
						return;
					}
					
					shipDestroyedTimer.Enabled = false;
					Console.SetCursorPosition (currentShipPosition.Left, currentShipPosition.Top);
					Thread.Sleep (1500);
					Console.Write ("   ");
					Thread.Sleep (1500);
					shipExplosionState = 0;
					DrawShip (ShipPosition.Initial);

					invadersBulletTimer.Enabled = true;
					invadersTimer.Enabled = true;
					shipBulletTimer.Enabled = true;
					shipBeingDestroyed = false;
					DrawReserveShips ();
				}
			}
		}

		private void OnInvadersExplosion (object source, ElapsedEventArgs e)
		{
			lock (this)
			{
				if (invaderExplosionState < 2)
				{
					Console.SetCursorPosition (invaderExplosion.Left, invaderExplosion.Top);
					if (invaderExplosionState == 0)
						Console.ForegroundColor = ConsoleColor.DarkYellow;
					else
						Console.ForegroundColor = ConsoleColor.Yellow;

					Console.Write ("{0}{0}{0}", (char)9788);
					invaderExplosionState++;
				}
				else
				{
					invaderExplosionState = 0;
					Console.MoveBufferArea(0, 0, 3, 1, invaderExplosion.Left, invaderExplosion.Top);
					explosionTimer.Enabled = false;
				}
			}
		}

		// Tracks movement, and collisions of invader bullets
		private void OnInvadersBulletTimedEvent (object source, ElapsedEventArgs e)
		{
			lock (this)
			{
				ResolveBulletCollisions ();
				ResolveShieldCollisions ();
				foreach (ItemPosition ipInvaderBullet in liInvaderBullets)
				{
					if (ipInvaderBullet.Top >= Console.WindowHeight - 2)
					{
						Console.MoveBufferArea (0, 0, 1, 1, ipInvaderBullet.Left, ipInvaderBullet.Top);
						liInvaderBullets.Remove (ipInvaderBullet);
					}
					else
					{
						Console.MoveBufferArea (ipInvaderBullet.Left, ipInvaderBullet.Top, 1, 1, ipInvaderBullet.Left, ipInvaderBullet.Top + 1);
						ipInvaderBullet.Top = ipInvaderBullet.Top + 1;
					}
				}

				GenerateInvaderBullet ();
				ResolveShipDestroyed ();
			}
		}

		// displays the high score table for a specific
		// period, and moves back to the main screen
		// when that period is up
		private void OnHighScoreTimedEvent(object source, ElapsedEventArgs e)
		{
			StartScreen();
			HighScoreTimer.Enabled = false;
		}
	}
}